#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>

#include <netinet/in.h>

#include <arpa/inet.h>

#include <unistd.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <fdevent.h>
#include <signal.h>
#include <ctype.h>

#include <stdio.h>

#include "server.h"
#include "keyvalue.h"
#include "log.h"
#include "connections.h"
#include "joblist.h"
#include "http_chunk.h"

#include "plugin.h"

typedef struct {
	char **ptr;
	
	size_t size;
	size_t used;
} char_array;

int cgi_pid_add(server *srv, pid_t pid) {
	int m = -1;
	size_t i;
	buffer_pid_t *r = &(srv->cgi_pid);
	
	for (i = 0; i < r->used; i++) {
		if (r->ptr[i] > m) m = r->ptr[i];
	}
	
	if (r->size == 0) {
		r->size = 16;
		r->ptr = malloc(sizeof(int) * r->size);
	} else if (r->used == r->size) {
		r->size += 16;
		r->ptr = realloc(r->ptr, sizeof(int) * r->size);
	}
	
	r->ptr[r->used++] = pid;
	
	return m;
}

int cgi_pid_del(server *srv, pid_t pid) {
	size_t i;
	buffer_pid_t *r = &(srv->cgi_pid);
	
	for (i = 0; i < r->used; i++) {
		if (r->ptr[i] == pid) break;
	}
	
	if (i != r->used) {
		/* found */
		
		if (i != r->used - 1) {
			r->ptr[i] = r->ptr[r->used - 1];
		}
		r->used--;
	}
	
	return 0;
}

int cgi_env_add(char_array *env, const char *key, const char *val) {
	int key_len, val_len;
	char *dst;
	
	if (!key || !val) return -1;
	
	key_len = strlen(key);
	val_len = strlen(val);
	
	dst = malloc(key_len + val_len + 3);
	memcpy(dst, key, key_len);
	dst[key_len] = '=';
	/* add the \0 from the value */
	memcpy(dst + key_len + 1, val, val_len + 1);
	
	if (env->size == 0) {
		env->size = 16;
		env->ptr = malloc(env->size * sizeof(char *));
	} else if (env->size == env->used) {
		env->size += 16;
		env->ptr = realloc(env->ptr, env->size * sizeof(char *));
	}
	
	env->ptr[env->used++] = dst;
	
	return 0;
}

int cgi_create_env(server *srv, connection *con, char *cgi_handler) {
	pid_t pid;
	
#ifdef HAVE_IPV6
	char b2[INET6_ADDRSTRLEN + 1];
#endif
	
	int to_cgi_fds[2];
	int from_cgi_fds[2];
	
	if (pipe(to_cgi_fds)) {
		log_error_write(srv, __FILE__, __LINE__, "ss", "pipe failed: ", strerror(errno));
		return -1;
	}
	
	if (pipe(from_cgi_fds)) {
		log_error_write(srv, __FILE__, __LINE__, "ss", "pipe failed: ", strerror(errno));
		return -1;
	}
	
	if (con->request.content->used) {
		write(to_cgi_fds[1], con->request.content->ptr, con->request.content_length);
	}
	
	/* fork, execve */
	switch (pid = fork()) {
	case 0: {
		/* child */
		char **args;
		int argc;
		int i = 0;
		char buf[32];
		size_t n;
		char_array env;
		char *c;
		
		/* move stdout to from_cgi_fd[1] */
		close(STDOUT_FILENO);
		dup(from_cgi_fds[1]);
		close(from_cgi_fds[1]);
		/* not needed */
		close(from_cgi_fds[0]);
		
		/* move the stdin to to_cgi_fd[0] */
		close(STDIN_FILENO);
		dup(to_cgi_fds[0]);
		close(to_cgi_fds[0]);
		/* not needed */
		close(to_cgi_fds[1]);
		
		/* create environment */
		env.ptr = NULL;
		env.size = 0;
		env.used = 0;
		
#ifdef PACKAGE_NAME
		cgi_env_add(&env, "SERVER_SOFTWARE", PACKAGE_NAME"/"PACKAGE_VERSION);
#else
		cgi_env_add(&env, "SERVER_SOFTWARE", PACKAGE"/"VERSION);
#endif
		cgi_env_add(&env, "SERVER_NAME", 
#ifdef HAVE_IPV6
			    inet_ntop(srv->srv_addr.plain.sa_family, 
				      srv->srv_addr.plain.sa_family == AF_INET6 ? 
				      (const void *) &(srv->srv_addr.ipv6.sin6_addr) :
				      (const void *) &(srv->srv_addr.ipv4.sin_addr),
				      b2, sizeof(b2)-1)
#else
			    inet_ntoa(srv->srv_addr.ipv4.sin_addr)
#endif
			    );
		cgi_env_add(&env, "GATEWAY_INTERFACE", "CGI/1.1");
		
		cgi_env_add(&env, "SERVER_PROTOCOL", get_http_version_name(con->request.http_version));
		
		ltostr(buf, 
#ifdef HAVE_IPV6
			ntohs(srv->srv_addr.plain.sa_family == AF_INET6 ? srv->srv_addr.ipv6.sin6_port : srv->srv_addr.ipv4.sin_port)
#else
			ntohs(srv->srv_addr.ipv4.sin_port)
#endif
			);
		cgi_env_add(&env, "SERVER_PORT", buf);
		cgi_env_add(&env, "REQUEST_METHOD", get_http_method_name(con->request.http_method));
		cgi_env_add(&env, "PATH_INFO", con->request.uri->ptr);
		cgi_env_add(&env, "REDIRECT_STATUS", "200");
		cgi_env_add(&env, "QUERY_STRING", con->uri.query->used ? con->uri.query->ptr : "");
		cgi_env_add(&env, "REMOTE_ADDR", 
#ifdef HAVE_IPV6
			    inet_ntop(con->dst_addr.plain.sa_family, 
				      con->dst_addr.plain.sa_family == AF_INET6 ? 
				      (const void *) &(con->dst_addr.ipv6.sin6_addr) :
				      (const void *) &(con->dst_addr.ipv4.sin_addr),
				      b2, sizeof(b2)-1)
#else
			    inet_ntoa(con->dst_addr.ipv4.sin_addr)
#endif
			    );
		
		ltostr(buf, con->request.content_length);
		cgi_env_add(&env, "CONTENT_LENGTH", buf);
		cgi_env_add(&env, "SCRIPT_FILENAME", con->physical.path->ptr);
		
		for (n = 0; n < con->request.headers->used; n++) {
			data_string *ds;
			
			ds = (data_string *)con->request.headers->data[n];
			
			if (ds->value->used && ds->key->used) {
				size_t j;
				
				buffer_reset(srv->tmp_buf);
				
				if (0 != strcasecmp(ds->key->ptr, "CONTENT_TYPE")) {
					buffer_copy_string(srv->tmp_buf, "HTTP_");
				}
				
				srv->tmp_buf->used--;
				buffer_prepare_append(srv->tmp_buf, ds->key->used + 1);
				
				for (j = 0; j < ds->key->used - 1; j++) {
					srv->tmp_buf->ptr[srv->tmp_buf->used++] = 
						isalpha((unsigned char)ds->key->ptr[j]) ? 
						toupper((unsigned char)ds->key->ptr[j]) : '_';
				}
				srv->tmp_buf->ptr[srv->tmp_buf->used++] = '\0';
				
				cgi_env_add(&env, srv->tmp_buf->ptr, ds->value->ptr);
			}
		}
		
		if (env.size == env.used) {
			env.size += 16;
			env.ptr = realloc(env.ptr, env.size * sizeof(char *));
		}
		
		env.ptr[env.used] = NULL;
		
		/* set up args */
		argc = 8;
		args = malloc(sizeof(char *) * argc);
		i = 0;
		
		args[i++] = cgi_handler;
		args[i++] = strdup(con->physical.path->ptr);
		
		args[i++] = NULL;

		/* search for the last / */
		if (NULL != (c = strrchr(con->physical.path->ptr, '/'))) {
			*c = '\0';
			
			/* change to the physical directory */
			if (-1 == chdir(con->physical.path->ptr)) {
				log_error_write(srv, __FILE__, __LINE__, "sss", "CGI failed: ", strerror(errno), con->physical.doc_root->ptr);
			}
		}

		/* we don't need the client socket */
		for (i = 3; i < 256; i++) {
			close(i);
		}
		
		/* exec the cgi */
		execve(args[0], args, env.ptr);
		
		log_error_write(srv, __FILE__, __LINE__, "sss", "CGI failed: ", strerror(errno), args[0]);
		
		exit(-1);
		break;
	}
	case -1:
		/* error */
		log_error_write(srv, __FILE__, __LINE__, "ss", "fork failed: ", strerror(errno));
		break;
	default:
		/* father */
		
		close(from_cgi_fds[1]);
		
		close(to_cgi_fds[0]);
		close(to_cgi_fds[1]);
		
		/* register PID and wait for them asyncronously */
		con->mode = CGI;
		con->cgi.pid = pid;
		con->cgi.fd = from_cgi_fds[0];
		buffer_reset(con->physical.path);
		
		con->cgi.fde_ndx = -1;
		
		fdevent_fdc_add(srv->ev->fdc, con->cgi.fd, con, FD_EVENT_TYPE_CGI_CONNECTION);
		con->cgi.fde_ndx = fdevent_event_add(srv->ev, con->cgi.fde_ndx, con->cgi.fd, FDEVENT_IN);
		
		if (-1 == fdevent_fcntl_set(srv->ev, con->cgi.fd)) {
			log_error_write(srv, __FILE__, __LINE__, "ss", "fcntl failed: ", strerror(errno));
			con->cgi.fde_ndx = fdevent_event_del(srv->ev, con->cgi.fde_ndx, con->cgi.fd);
			fdevent_fdc_del(srv->ev->fdc, con->cgi.fd);
			close(con->cgi.fd);
			return -1;
		}
		
		break;
	}
	
	return 0;
}

handler_t cgi_is_handled(server *srv, connection *con, void *p_d) {
	size_t k, s_len;
	buffer *fn = con->physical.path;
	
	if (fn->used == 0) return HANDLER_ERROR;
	
	s_len = fn->used - 1;
	
	for (k = 0; k < srv->conf.cgi->used; k++) {
		data_string *ds = (data_string *)srv->conf.cgi->data[k];
		size_t ct_len = ds->key->used - 1;
		
		if (ds->key->used == 0) continue;
		if (s_len < ct_len) continue;
		
		if (0 == strncmp(fn->ptr + s_len - ct_len, ds->key->ptr, ct_len)) {
			cgi_create_env(srv, con, ds->value->ptr);
			
			return HANDLER_FINISHED;
		}
	}
	
	return HANDLER_GO_ON;
}

enum {FDEVENT_HANDLED_UNSET, FDEVENT_HANDLED_FINISHED, FDEVENT_HANDLED_NOT_FINISHED, FDEVENT_HANDLED_ERROR};

int cgi_response_parse(server *srv, connection *con, buffer *in) {
	char *s, *ns;
	
	/* \r\n -> \0\0 */
	
	buffer_copy_string(srv->parse_response, in->ptr);
	
	for (s = srv->parse_response->ptr; NULL != (ns = strstr(s, "\r\n")); s = ns + 2) {
		char *key, *value;
		int key_len;
		data_string *ds;
		
		ns[0] = '\0';
		ns[1] = '\0';
		
		key = s;
		if (NULL == (value = strstr(s, ": "))) {
			/* we expect: "<key>: <value>\n" */
			continue;
		}
		
		key_len = value - key;
		value += 2;
		
		if (NULL == (ds = (data_string *)array_get_unused_element(con->response.headers, TYPE_STRING))) {
			ds = data_string_init();
		}
		buffer_copy_string_len(ds->key, key, key_len);
		buffer_copy_string(ds->value, value);
		
		array_insert_unique(con->response.headers, (data_unset *)ds);
		
		switch(key_len) {
		case 4:
			if (0 == strncasecmp(key, "Date", key_len)) {
				con->parsed_response |= HTTP_DATE;
			}
			break;
		case 6:
			if (0 == strncasecmp(key, "Status", key_len)) {
				con->http_status = strtol(value, NULL, 10);
				con->parsed_response |= HTTP_STATUS;
			}
			break;
		case 8:
			if (0 == strncasecmp(key, "Location", key_len)) {
				con->parsed_response |= HTTP_LOCATION;
			}
			break;
		case 10:
			if (0 == strncasecmp(key, "Connection", key_len)) {
				con->response.keep_alive = (0 == strcasecmp(value, "Keep-Alive")) ? TRUE : FALSE;
				con->parsed_response |= HTTP_CONNECTION;
			}
			break;
		case 14:
			if (0 == strncasecmp(key, "Content-Length", key_len)) {
				con->response.content_length = strtol(value, NULL, 10);
				con->parsed_response |= HTTP_CONTENT_LENGTH;
			}
			break;
		default:
			break;
		}
	}
	
	/* CGI/1.1 rev 03 - 7.2.1.2 */
	if ((con->parsed_response & HTTP_LOCATION) &&
	    !(con->parsed_response & HTTP_STATUS)) {
		con->http_status = 302;
	}
	
	return 0;
}


int cgi_demux_response(server *srv, connection *con) {
	while(1) {
		int n;
		
		buffer_prepare_copy(con->cgi.response, 1024);
		if (-1 == (n = read(con->cgi.fd, con->cgi.response->ptr, con->cgi.response->size - 1))) {
			if (errno == EAGAIN) {
				/* would block, wait for signal */
				return FDEVENT_HANDLED_NOT_FINISHED;
			}
			/* error */
			log_error_write(srv, __FILE__, __LINE__, "sdd", strerror(errno), con->fd, con->cgi.fd);
			return FDEVENT_HANDLED_ERROR;
		}
		
		if (n == 0) {
			/* read finished */
			
			con->file_finished = 1;
			
			/* send final chunk */
			http_chunk_append_mem(srv, con, NULL, 0);
			joblist_append(srv, con);
			
			return FDEVENT_HANDLED_FINISHED;
		}
		
		con->cgi.response->ptr[n] = '\0';
		con->cgi.response->used = n+1;
		
		/* split header from body */
		
		if (con->file_started == 0) {
			char *c;
			
			/* search for the \r\n\r\n in the string */
			if (NULL != (c = buffer_search_string_len(con->cgi.response, "\r\n\r\n", 4))) {
				size_t hlen = c - con->cgi.response->ptr + 4;
				size_t blen = con->cgi.response->used - hlen - 1;
				/* found */
				
				buffer_append_string_len(con->cgi.response_header, con->cgi.response->ptr, c - con->cgi.response->ptr + 4);
#if 0
				log_error_write(srv, __FILE__, __LINE__, "sddd", "Header-len:", hlen, blen, con->cgi.response->used);
#endif
				/* parse the response header */
				cgi_response_parse(srv, con, con->cgi.response_header);
				
				/* enable chunked-transfer-encoding */
				if (con->request.http_version == HTTP_VERSION_1_1 &&
				    !(con->parsed_response & HTTP_CONTENT_LENGTH)) {
					con->response.transfer_encoding = HTTP_TRANSFER_ENCODING_CHUNKED;
				}
				
				con->file_started = 1;
				
				if ((con->cgi.response->used != hlen) &&
				    blen) {
					http_chunk_append_mem(srv, con, c + 4, blen + 1);
					joblist_append(srv, con);
					
#if 0
					log_error_write(srv, __FILE__, __LINE__, "sd", "body-len", blen);
#endif
				}
			} else {
				/* copy everything */
				buffer_append_string(con->cgi.response_header, con->cgi.response->ptr);
			}
		} else {
#if 0
			log_error_write(srv, __FILE__, __LINE__, "sd", "body-len", con->cgi.response->used - 1);
#endif
			
			http_chunk_append_mem(srv, con, con->cgi.response->ptr, con->cgi.response->used);
			joblist_append(srv, con);

		}
		
#if 0		
		log_error_write(srv, __FILE__, __LINE__, "ddss", con->fd, con->cgi.fd, connection_get_state(con->state), b->ptr);
#endif
	}
	
	return FDEVENT_HANDLED_NOT_FINISHED;
}

handler_t cgi_connection_close(server *srv, connection *con, void *p_d) {
	int status, r = 0;
	
	if (con->mode != CGI) return HANDLER_GO_ON;
	
#if 0	
	log_error_write(srv, __FILE__, __LINE__, "sdd", 
			"emergency exit: cgi", 
			con->fd,
			con->cgi.fd);
#endif
	
	/* the connection to the browser went away, but we still have a connection
	 * to the CGI script 
	 *
	 * close cgi-connection
	 */
	
	/* close connection to the cgi-script */
	con->cgi.fde_ndx = fdevent_event_del(srv->ev, con->cgi.fde_ndx, con->cgi.fd);
	fdevent_fdc_del(srv->ev->fdc, con->cgi.fd);
	close(con->cgi.fd);
	con->cgi.fd = -1;
	con->cgi.fde_ndx = -1;
	
	/* if waitpid hasn't been called by response.c yet, do it here */
	if (con->cgi.pid) {
		/* check if the CGI-script is already gone */
		switch(r = waitpid(con->cgi.pid, &status, WNOHANG)) {
		case 0:
			/* not finished yet */
			log_error_write(srv, __FILE__, __LINE__, "sd", "child isn't done yet, fd:", con->fd);
			break;
		case -1:
			log_error_write(srv, __FILE__, __LINE__, "ss", "waitpid failed: ", strerror(errno));
			return HANDLER_ERROR;
		default:
			if (WIFEXITED(status)) {
#if 0
				log_error_write(srv, __FILE__, __LINE__, "s", "cgi exited fine");
#endif
				con->cgi.pid = 0;
				return HANDLER_FINISHED;
			} else {
				log_error_write(srv, __FILE__, __LINE__, "s", "cgi died ?");
				con->cgi.pid = 0;
				return HANDLER_FINISHED;
			}
		}
		
	
		kill(con->cgi.pid, SIGTERM);
		
		/* cgi-script is still alive, queue the PID for removal */
		cgi_pid_add(srv, con->cgi.pid);
	}
		
	return HANDLER_FINISHED;
}

handler_t cgi_handle_fdevent(server *srv, const fd_conn *fdc, void *p_d) {
	connection *con = fdc->conn;
	int revents = fdc->revents;
	
	if (fdc->fd_type != FD_EVENT_TYPE_CGI_CONNECTION) return HANDLER_GO_ON;
	
	joblist_append(srv, con);
	
	if (con->cgi.fd == -1) {
		log_error_write(srv, __FILE__, __LINE__, "ddss", con->fd, con->cgi.fd, connection_get_state(con->state), "invalid cgi-fd");
		
		return HANDLER_ERROR;
	}
	
	if ((fdc->revents & FDEVENT_IN)) {
		switch (cgi_demux_response(srv, con)) {
		case FDEVENT_HANDLED_NOT_FINISHED:
			break;
		case FDEVENT_HANDLED_FINISHED:
			/* we are done */
			
#if 0
			log_error_write(srv, __FILE__, __LINE__, "ddss", con->fd, con->cgi.fd, connection_get_state(con->state), "finished");
#endif
			
			break;
		case FDEVENT_HANDLED_ERROR:
			connection_set_state(srv, con, HANDLE_REQUEST);
			con->http_status = 500;
			con->mode = DIRECT;
			
			log_error_write(srv, __FILE__, __LINE__, "s", "demuxer failed: ");
			break;
		}
	}
	
	if (fdc->revents & FDEVENT_OUT) {
		/* nothing to do */
	}
	
	/* perhaps this issue is already handled */
	if (fdc->revents & FDEVENT_HUP) {
		if (con->file_finished == 0) {
			http_chunk_append_mem(srv, con, NULL, 0);
			joblist_append(srv, con);
		}
		
		con->file_finished = 1;
		
		if (chunkqueue_is_empty(con->write_queue)) {
			/* there is nothing left to write */
			connection_set_state(srv, con, RESPONSE_END);
		} else {
			/* used the write-handler to finish the request on demand */
			
		}
		
# if 0
		log_error_write(srv, __FILE__, __LINE__, "sddd", "got HUP from cgi", con->fd, con->cgi.fd, revents);
# endif
		
		/* rtsigs didn't liked the close */
		cgi_connection_close(srv, con, p_d);
	} else if (revents & FDEVENT_ERR) {
		con->file_finished = 1;
		
		/* kill all connections to the cgi process */
		cgi_connection_close(srv, con, p_d);
#if 1
		log_error_write(srv, __FILE__, __LINE__, "s", "cgi-FDEVENT_ERR");
#endif			
		return HANDLER_ERROR;
	}
	
	return HANDLER_FINISHED;
}

TRIGGER_FUNC(cgi_trigger) {
	size_t ndx;
	
	for (ndx = 0; ndx < srv->cgi_pid.used; ndx++) {
		int status;
		switch(waitpid(srv->cgi_pid.ptr[ndx], &status, WNOHANG)) {
		case 0:
			/* not finished yet */
			log_error_write(srv, __FILE__, __LINE__, "sd", "child isn't done yet, pid:", srv->cgi_pid.ptr[ndx]);
			break;
		case -1:
			log_error_write(srv, __FILE__, __LINE__, "ss", "waitpid failed: ", strerror(errno));
			return HANDLER_ERROR;
		default:
			if (WIFEXITED(status)) {
				log_error_write(srv, __FILE__, __LINE__, "s", "cgi exited fine");
			} else {
				log_error_write(srv, __FILE__, __LINE__, "s", "cgi died ?");
			}
			
			cgi_pid_del(srv, srv->cgi_pid.ptr[ndx]);
			/* del modified the buffer structure 
			 * and copies the last entry to the current one
			 * -> recheck the current index
			 */
			ndx--;
		}
	}
	
	return HANDLER_GO_ON;
}

SUBREQUEST_FUNC(mod_cgi_handle_subrequest) {
	int status;
	
	if (con->mode != CGI) return HANDLER_GO_ON;
	
	if (con->cgi.pid == 0) return HANDLER_FINISHED;
	
	switch(waitpid(con->cgi.pid, &status, WNOHANG)) {
	case 0:
		/* not finished yet */
		if (con->file_started) {
			return HANDLER_FINISHED;
		} else {
			return HANDLER_WAIT_FOR_EVENT;
		}
	case -1:
		log_error_write(srv, __FILE__, __LINE__, "ss", "waitpid failed: ", strerror(errno));
		con->mode = DIRECT;
		con->http_status = 500;
		return HANDLER_FINISHED;
	default:
		if (WIFEXITED(status)) {
			con->cgi.pid = 0;
				return HANDLER_FINISHED;
		} else {
			log_error_write(srv, __FILE__, __LINE__, "s", "cgi died ?");
			con->cgi.pid = 0;
			
			con->mode = DIRECT;
			con->http_status = 500;
			return HANDLER_FINISHED;
		}
	}
}


int mod_cgi_plugin_init(plugin *p) {
	p->name        = buffer_init_string("cgi");

	p->handle_connection_close = cgi_connection_close;
	p->handle_subrequest_start = cgi_is_handled;
	p->handle_subrequest = mod_cgi_handle_subrequest;
	p->handle_fdevent = cgi_handle_fdevent;
	p->handle_trigger = cgi_trigger;
	
	p->data        = NULL;
	
	return 0;
}
