#include <sys/types.h>
#include <sys/mman.h>
#include <sys/stat.h>

#include <fcntl.h>
#include <unistd.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <time.h>

#include <zlib.h>

#include "base.h"
#include "log.h"
#include "buffer.h"
#include "response.h"

#include "plugin.h"

#include "crc32.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

typedef struct {
	buffer *ofn;
	buffer *b;
	
	buffer *conf_compress_cache_dir;
	
	array  *conf_compress;
} plugin_data;

INIT_FUNC(mod_compress_init) {
	plugin_data *p;
	
	p = malloc(sizeof(plugin_data));
	memset(p, 0, sizeof(plugin_data));
	
	p->ofn = buffer_init();
	p->b = buffer_init();
	p->conf_compress_cache_dir = buffer_init();
	
	p->conf_compress = array_init();
	
	return p;
}

FREE_FUNC(mod_compress_free) {
	plugin_data *p = p_d;
	
	if (!p) return HANDLER_GO_ON;
	
	buffer_free(p->ofn);
	buffer_free(p->b);
	buffer_free(p->conf_compress_cache_dir);
	
	array_free(p->conf_compress);
	
	free(p);
	
	return HANDLER_GO_ON;
}

SETDEFAULTS_FUNC(mod_compress_setdefaults) {
	plugin_data *p = p_d;
	struct stat st;
	
	config_values_t cv[] = { 
		{ "compress.cache-dir",             NULL, T_CONFIG_STRING },
		{ "compress.filetype",              NULL, T_CONFIG_ARRAY },
		{ NULL,                             NULL, T_CONFIG_UNSET }
	};
	
	cv[0].destination = p->conf_compress_cache_dir;
	cv[1].destination = p->conf_compress;
	
	if (0 != config_insert_values(srv, cv)) {
		return HANDLER_ERROR;
	}
	
	if (p->conf_compress_cache_dir->used == 0) {
		log_error_write(srv, __FILE__, __LINE__, "s", "compress.cache-dir has to be set");
		
		return HANDLER_ERROR;
	}
	
	if (0 != stat(p->conf_compress_cache_dir->ptr, &st)) {
		log_error_write(srv, __FILE__, __LINE__, "sbs", "can't stat compress.cache-dir", p->conf_compress_cache_dir, strerror(errno));
		
		return HANDLER_ERROR;
	}
	
	return HANDLER_GO_ON;
}


int deflate_file(server *srv, connection *con, plugin_data *p, buffer *fn, file_cache_entry *fce, int type) {
	z_stream z;
	int ifd, ofd;
	int ret;
	void *start;
	const char *filename = fn->ptr;
	
	/* overflow */
	if ((off_t)(fce->st.st_size * 1.1) < fce->st.st_size) return -1;
	
	buffer_reset(p->ofn);
	buffer_copy_string_buffer(p->ofn, p->conf_compress_cache_dir);
	BUFFER_APPEND_SLASH(p->ofn);
	
	if (0 == strncmp(con->physical.path->ptr, con->physical.doc_root->ptr, con->physical.doc_root->used-1)) {
		size_t offset = p->ofn->used - 1;
		char *dir, *nextdir;
		
		buffer_append_string(p->ofn, con->physical.path->ptr + con->physical.doc_root->used - 1);
		
		buffer_copy_string_buffer(p->b, p->ofn);
		
		/* mkdir -p ... */
		for (dir = p->b->ptr + offset; NULL != (nextdir = strchr(dir, '/')); dir = nextdir + 1) {
			*nextdir = '\0';
			
			if (-1 == mkdir(p->b->ptr, 0700)) {
				if (errno != EEXIST) {
					log_error_write(srv, __FILE__, __LINE__, "ssss", "creating cache-directory", p->b->ptr, "failed", strerror(errno));
					
					return -1;
				}
			}
			
			*nextdir = '/';
		}
	} else {
		buffer_append_string_buffer(p->ofn, con->uri.path);
	}
	
	if (type == HTTP_ACCEPT_ENCODING_GZIP) {
		buffer_append_string(p->ofn, "-gzip-");
	} else if (type == HTTP_ACCEPT_ENCODING_DEFLATE) {
		buffer_append_string(p->ofn, "-deflate-");
	} else {
		log_error_write(srv, __FILE__, __LINE__, "sd", "unknown compression type", type);
		return -1;
	}
	
	buffer_append_string_buffer(p->ofn, fce->etag);
	
	if (-1 == (ofd = open(p->ofn->ptr, O_WRONLY | O_CREAT | O_EXCL, 0600))) {
		if (errno == EEXIST) {
			/* cache-entry exists */
#if 0
			log_error_write(srv, __FILE__, __LINE__, "bs", p->ofn, "compress-cache hit");
#endif
			buffer_copy_string_buffer(con->physical.path, p->ofn);
			
			return 0;
		}
		
		log_error_write(srv, __FILE__, __LINE__, "sbss", "creating cachefile", p->ofn, "failed", strerror(errno));
		
		return -1;
	}
#if 0
	log_error_write(srv, __FILE__, __LINE__, "bs", p->ofn, "compress-cache miss");
#endif	
	if (-1 == (ifd = open(filename, O_RDONLY))) {
		log_error_write(srv, __FILE__, __LINE__, "sbss", "opening plain-file", fn, "failed", strerror(errno));
		
		close(ofd);
		
		return -1;
	}
	
	
	if (MAP_FAILED == (start = mmap(NULL, fce->st.st_size, PROT_READ, MAP_SHARED, ifd, 0))) {
		log_error_write(srv, __FILE__, __LINE__, "sbss", "mmaping", fn, "failed", strerror(errno));
		
		close(ofd);
		close(ifd);
		return -1;
	}
	
	z.zalloc = Z_NULL;
	z.zfree = Z_NULL;
	z.opaque = Z_NULL;
	
	if (Z_OK != deflateInit2(&z, 
				 Z_DEFAULT_COMPRESSION,
				 Z_DEFLATED, 
				 -MAX_WBITS,  /* supress zlib-header */
				 8,
				 Z_DEFAULT_STRATEGY)) {
		munmap(start, fce->st.st_size);
		close(ofd);
		close(ifd);
		return -1;
	}
	
	z.next_in = start;
	z.avail_in = fce->st.st_size;
	z.total_in = 0;
	
	if (type == HTTP_ACCEPT_ENCODING_GZIP) {
		unsigned char *c;
		buffer_prepare_copy(p->b, (z.avail_in * 1.1) + 12 + 18);
		
		/* write gzip header */
		
		c = (unsigned char *)p->b->ptr;
		c[0] = 0x1f;
		c[1] = 0x8b;
		c[2] = Z_DEFLATED;
		c[3] = 0; /* options */
		c[4] = (fce->st.st_mtime >>  0) & 0xff;
		c[5] = (fce->st.st_mtime >>  8) & 0xff;
		c[6] = (fce->st.st_mtime >> 16) & 0xff;
		c[7] = (fce->st.st_mtime >> 24) & 0xff;
		c[8] = 0x00; /* extra flags */
		c[9] = 0x03; /* UNIX */
	
		p->b->used = 10;
		z.next_out = (unsigned char *)p->b->ptr + p->b->used;
		z.avail_out = p->b->size - p->b->used - 8;
		z.total_out = 0;
	} else {
		buffer_prepare_copy(p->b, (z.avail_in * 1.1) + 12);
	
		z.next_out = (unsigned char *)p->b->ptr;
		z.avail_out = p->b->size;
		z.total_out = 0;
	}
	
	if (Z_STREAM_END != (ret = deflate(&z, Z_FINISH))) {
		deflateEnd(&z);
		munmap(start, fce->st.st_size);
		close(ofd);
		close(ifd);
		return -1;
	}
	
	/* trailer */
	p->b->used += z.total_out;
	
	if (HTTP_ACCEPT_ENCODING_GZIP == type) {
		unsigned char *c;
		unsigned long crc = generate_crc32c(start, fce->st.st_size);
		
		c = (unsigned char *)p->b->ptr + p->b->used;
		
		c[0] = (crc >>  0) & 0xff;
		c[1] = (crc >>  8) & 0xff;
		c[2] = (crc >> 16) & 0xff;
		c[3] = (crc >> 24) & 0xff;
		c[4] = (z.total_in >>  0) & 0xff;
		c[5] = (z.total_in >>  8) & 0xff;
		c[6] = (z.total_in >> 16) & 0xff;
		c[7] = (z.total_in >> 24) & 0xff;
		p->b->used += 8;
	}

	if (p->b->used != write(ofd, p->b->ptr, p->b->used)) {
		munmap(start, fce->st.st_size);
		close(ofd);
		close(ifd);
		return -1;
	}
	
	if (Z_OK != deflateEnd(&z)) {
		munmap(start, fce->st.st_size);
		close(ofd);
		close(ifd);
		
		return -1;
	}
	munmap(start, fce->st.st_size);
	close(ofd);
	close(ifd);
	
	buffer_copy_string_buffer(con->physical.path, p->ofn);
	
	return 0;
}

PHYSICALPATH_FUNC(mod_compress_physical) {
	plugin_data *p = p_d;
	
	if (con->request.http_method == HTTP_METHOD_GET || 
	    con->request.http_method == HTTP_METHOD_POST) {
		data_string *content_ds;
		size_t m;
		
		content_ds = (data_string *)array_get_element(con->response.headers, "Content-Type");
		
		if (!content_ds) {
			log_error_write(srv, __FILE__, __LINE__, "sbb", "Content-Type is not set for", con->physical.path, con->uri.path);
			
			return HANDLER_GO_ON;
		}
		
		/* check if mimetype is in compress-config */
		for (m = 0; m < p->conf_compress->used; m++) {
			data_string *compress_ds = (data_string *)p->conf_compress->data[m];
			
			if (!compress_ds) {
				log_error_write(srv, __FILE__, __LINE__, "sbb", "evil", con->physical.path, con->uri.path);
				
				return HANDLER_GO_ON;
			}
			
			if (buffer_is_equal(compress_ds->value, content_ds->value)) {
				/* mimetype found */
				data_string *ds;
				
				/* the response might change according to Accept-Encoding */
				response_header_insert(srv, con, "Vary", "Accept-Encoding");
				
				
				if (NULL != (ds = (data_string *)array_get_element(con->request.headers, "Accept-Encoding"))) {
					int accept_encoding = 0;
					char *value = ds->value->ptr;
			
					/* parse the possible encodings */
					if (NULL != strstr(value, "gzip")) accept_encoding |= HTTP_ACCEPT_ENCODING_GZIP;
					if (NULL != strstr(value, "deflate")) accept_encoding |= HTTP_ACCEPT_ENCODING_DEFLATE;
					if (NULL != strstr(value, "compress")) accept_encoding |= HTTP_ACCEPT_ENCODING_COMPRESS;
					if (NULL != strstr(value, "identity")) accept_encoding |= HTTP_ACCEPT_ENCODING_IDENTITY;
					
					if (accept_encoding & (HTTP_ACCEPT_ENCODING_DEFLATE | HTTP_ACCEPT_ENCODING_GZIP)) {
						/* deflate it */
				
						if (0 == deflate_file(srv, con, p,
								      con->physical.path, con->fce, 
								      (accept_encoding & HTTP_ACCEPT_ENCODING_GZIP) ? 
								      HTTP_ACCEPT_ENCODING_GZIP : 
								      HTTP_ACCEPT_ENCODING_DEFLATE)) {
							size_t i;
							struct tm *tm;
							time_t last_mod;
							
							response_header_insert(srv, con, "Content-Encoding", 
									       (accept_encoding & HTTP_ACCEPT_ENCODING_GZIP) ?
									       "gzip" : "deflate");
							
							/* Set Last-Modified of ORIGINAL file */
							last_mod = con->fce->st.st_mtime;
							
							for (i = 0; i < FILE_CACHE_MAX; i++) {
								if (srv->mtime_cache[i].mtime == last_mod) break;
								
								if (srv->mtime_cache[i].mtime == 0) {
									srv->mtime_cache[i].mtime = last_mod;
									
									buffer_prepare_copy(srv->mtime_cache[i].str, 1024);
									
									tm = gmtime(&(srv->mtime_cache[i].mtime));
									srv->mtime_cache[i].str->used = strftime(srv->mtime_cache[i].str->ptr, 
														 srv->mtime_cache[i].str->size - 1,
														 "%a, %d %b %Y %H:%M:%S GMT", tm);
									
									srv->mtime_cache[i].str->used++;
									break;
								}
							}
							
							if (i == FILE_CACHE_MAX) {
								i = 0;
								
								srv->mtime_cache[i].mtime = last_mod;
								buffer_prepare_copy(srv->mtime_cache[i].str, 1024);
								tm = gmtime(&(srv->mtime_cache[i].mtime));
								srv->mtime_cache[i].str->used = strftime(srv->mtime_cache[i].str->ptr, 
													 srv->mtime_cache[i].str->size - 1,
													 "%a, %d %b %Y %H:%M:%S GMT", tm);
								srv->mtime_cache[i].str->used++;
							}
							
							response_header_insert(srv, con, "Last-Modified", srv->mtime_cache[i].str->ptr);
						
							return HANDLER_FINISHED;
						}
						break;
					}
				}
			}
		}
	}
	
	return HANDLER_GO_ON;
}

int mod_compress_plugin_init(plugin *p) {
	p->name        = buffer_init_string("compress");
	
	p->init        = mod_compress_init;
	p->set_defaults = mod_compress_setdefaults;
	p->handle_physical_path  = mod_compress_physical;
	p->cleanup     = mod_compress_free;
	
	p->data        = NULL;
	
	return 0;
}
