#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/resource.h>

#include <netinet/in.h>
#include <netinet/tcp.h>

#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <netdb.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>

#include "network.h"
#include "fdevent.h"
#include "log.h"
#include "file_cache.h"

#include "network_backends.h"

int network_server_init(server *srv, const char *host, unsigned short port) {
	int val;
	socklen_t addr_len;
	struct rlimit rlim;
	
#ifdef SO_ACCEPTFILTER
	struct accept_filter_arg afa;
#endif
	
#ifdef HAVE_IPV6
	if (srv->srvconf.use_ipv6) srv->srv_addr.plain.sa_family = AF_INET6;
	else srv->srv_addr.plain.sa_family = AF_INET;
	
	if (-1 == (srv->fd = socket(srv->srv_addr.plain.sa_family, SOCK_STREAM, IPPROTO_TCP))) {
		
		/* fallback to IPv4 */
		srv->srv_addr.plain.sa_family = AF_INET;
		if (-1 == (srv->fd = socket(srv->srv_addr.plain.sa_family, SOCK_STREAM, IPPROTO_TCP))) {
			log_error_write(srv, __FILE__, __LINE__, "ss", "socket failed:", strerror(errno));
			return -1;
		}
	}
#else
	srv->srv_addr.plain.sa_family = AF_INET;
	if (-1 == (srv->fd = socket(srv->srv_addr.plain.sa_family, SOCK_STREAM, IPPROTO_TCP))) {
		log_error_write(srv, __FILE__, __LINE__, "ss", "socket failed:", strerror(errno));
		return -1;
	}
#endif
	val = 1;
	if (setsockopt(srv->fd, SOL_SOCKET, SO_REUSEADDR, &val, sizeof val) < 0) {
		log_error_write(srv, __FILE__, __LINE__, "ss", "socketsockopt failed:", strerror(errno));
		return -1;
	}
	
	/* set rlimits */
	if (0 != getrlimit(RLIMIT_NOFILE, &rlim)) {
		log_error_write(srv, __FILE__, __LINE__,
				"ss", "couldn't get 'max filedescriptors'",
				strerror(errno));
		return -1;
	}
	
	if (NULL == (srv->ev = fdevent_init(rlim.rlim_cur, srv->event_handler))) {
		log_error_write(srv, __FILE__, __LINE__,
				"s", "fdevent_init failed");
		return -1;
	}
	
	if (-1 == fdevent_fcntl_set(srv->ev, srv->fd)) {
		log_error_write(srv, __FILE__, __LINE__, "ss", "fcntl failed:", strerror(errno));
		return -1;
	}
	
	switch(srv->srv_addr.plain.sa_family) {
#ifdef HAVE_IPV6
	case AF_INET6:
		memset(&srv->srv_addr, 0, sizeof(struct sockaddr_in6));
		srv->srv_addr.ipv6.sin6_family = AF_INET6;
		if (host == NULL) {
			srv->srv_addr.ipv6.sin6_addr = in6addr_any;
		} else {
			struct addrinfo hints, *res;
			int r;
			
			memset(&hints, 0, sizeof(hints));
			
			hints.ai_family   = AF_INET6;
			hints.ai_socktype = SOCK_STREAM;
			hints.ai_protocol = IPPROTO_TCP;
			
			if (0 != (r = getaddrinfo(host, NULL, &hints, &res))) {
				log_error_write(srv, __FILE__, __LINE__, 
						"sssss", "getaddrinfo failed: ", 
						gai_strerror(r), "'", host, "'");
				
				return -1;
			}
			
			memcpy(&(srv->srv_addr), res->ai_addr, res->ai_addrlen);
			
			freeaddrinfo(res);
		}
		srv->srv_addr.ipv6.sin6_port = htons(port);
		addr_len = sizeof(struct sockaddr_in6);
		break;
#endif
	case AF_INET:
		memset(&srv->srv_addr, 0, sizeof(struct sockaddr_in));
		srv->srv_addr.ipv4.sin_family = AF_INET;
		if (host == NULL) {
			srv->srv_addr.ipv4.sin_addr.s_addr = htonl(INADDR_ANY);
		} else {
			struct hostent *he;
			if (NULL == (he = gethostbyname(host))) {
				log_error_write(srv, __FILE__, __LINE__, 
						"sss", "gethostbyname failed: ", 
						hstrerror(h_errno), host);
				return -1;
			}
			
			if (he->h_addrtype != AF_INET) {
				log_error_write(srv, __FILE__, __LINE__, "sd", "addr-type != AF_INET: ", he->h_addrtype);
				return -1;
			}
			
			if (he->h_length != sizeof(struct in_addr)) {
				log_error_write(srv, __FILE__, __LINE__, "sd", "addr-length != sizeof(in_addr): ", he->h_length);
				return -1;
			}
			
			memcpy(&(srv->srv_addr.ipv4.sin_addr.s_addr), he->h_addr_list[0], he->h_length);
		}
		srv->srv_addr.ipv4.sin_port = htons(port);
		
		addr_len = sizeof(struct sockaddr_in);
		
		break;
	default:
		addr_len = 0;
		
		return -1;
	}
	
	if (0 != bind(srv->fd, (struct sockaddr *) &(srv->srv_addr), addr_len)) {
		log_error_write(srv, __FILE__, __LINE__, "ss", "can't bind: ", strerror(errno));
		return -1;
	}
	
	if (-1 == listen(srv->fd, 128 * 8)) {
		log_error_write(srv, __FILE__, __LINE__, "ss", "listen failed: ", strerror(errno));
		return -1;
	}
#ifdef SO_ACCEPTFILTER
	/*
	 * FreeBSD accf_http filter
	 *
	 */
	bzero(&afa, sizeof(afa));
	strcpy(afa.af_name, "httpready");
	if (setsockopt(srv->fd, SOL_SOCKET, SO_ACCEPTFILTER, &afa, sizeof(afa)) < 0) {
		log_error_write(srv, __FILE__, __LINE__, "ssd", "(notice) can't set accept-filter 'httpready': ", strerror(errno), errno);
	}
#endif
	
	return 0;
}


int network_server_close(server *srv) {
	close(srv->fd);
	
	return 0;
}

int network_write_chunkqueue(server *srv, connection *con, chunkqueue *cq) {
	int ret = -1, i;
#ifdef TCP_CORK	
	i = 1;
	setsockopt(con->fd, IPPROTO_TCP, TCP_CORK, &i, sizeof(i));
#endif
	
	if (srv->is_ssl) {
#ifdef USE_OPENSSL
		ret = network_write_chunkqueue_openssl(srv, con, cq);
#endif
	} else {
		/* dispatch call */
#if defined USE_LINUX_SENDFILE
		ret = network_write_chunkqueue_linuxsendfile(srv, con, cq); 
#elif defined USE_FREEBSD_SENDFILE
		ret = network_write_chunkqueue_freebsdsendfile(srv, con, cq); 
#elif defined USE_SOLARIS_SENDFILEV
		ret = network_write_chunkqueue_solarissendfilev(srv, con, cq); 
#elif defined USE_WRITEV
		ret = network_write_chunkqueue_writev(srv, con, cq);
#else
		ret = network_write_chunkqueue_write(srv, con, cq);
#endif
	}
	
	if (ret >= 0) {
		/*
		 * map the return code
		 * 
		 * -1 -> -1
		 * >0 -> (everything written) 0
		 *       (not finished yet)   1
		 * 
		 * ret means:
		 * - <ret> chunks are unused now
		 * 
		 */
		
		chunk *c, *pc = NULL;
		
		for (i = 0, c = cq->first; i < ret; i++, c = c->next) {
			if (i == ret - 1) pc = c;
		}
		
		if (c) {
			/* there is still something to write */
			
			if (c != cq->first) {
				/* move the first few buffers to unused */
				
				assert(pc);
				
				pc->next = cq->unused;
				cq->unused = cq->first;
				cq->first = c;
			}
			
			ret = 1;
		} else {
			/* everything is written */
			chunkqueue_reset(cq);
			
			ret = 0;
		}
	}
		
#ifdef TCP_CORK
	i = 0;
	setsockopt(con->fd, IPPROTO_TCP, TCP_CORK, &i, sizeof(i));
#endif
	return ret;
}
